@@ -21,20 +21,24 @@ Methods to annotate the Abstract Syntax Tree (AST) of an IDP-Z3 program.
 """
 
 from copy import copy
+from itertools import chain
+import string
+from typing import Dict
 
-from .Parse import (Vocabulary, Extern, TypeDeclaration,
-                    SymbolDeclaration, Symbol,
-                    Theory, Definition, Rule,
+from .Parse import (Vocabulary, Import, TypeDeclaration, Declaration, Type,
+                    SymbolDeclaration, Symbol, VarDeclaration,
+                    TheoryBlock, Definition, Rule,
                     Structure, SymbolInterpretation, Enumeration, FunctionEnum,
                     Tuple, ConstructedFrom, Display)
-from .Expression import (Expression, Constructor, IfExpr, AQuantification, Quantee,
-                         ARImplication, AImplication, AEquivalence, ADisjunction,
-                         AConjunction, Operator, AComparison, AUnary, AAggregate,
+from .Expression import (Expression, Constructor, AIfExpr, AQuantification, Quantee,
+                         ARImplication, AImplication, AEquivalence,
+                         Operator, AComparison, AUnary, AAggregate,
                          AppliedSymbol, UnappliedSymbol, Variable, Brackets,
-                         FALSE, SymbolExpr, Number)
+                         FALSE, SymbolExpr, Number, NOT, EQUALS, AND, OR, TRUE, FALSE,
+                         IMPLIES, RIMPLIES, EQUIV, FORALL, EXISTS, Extension)
 
-from .utils import (BOOL, INT, REAL, DATE, SYMBOL, OrderedSet, IDPZ3Error,
-                    DEF_SEMANTICS, Semantics)
+from .utils import (BOOL, INT, REAL, DATE, CONCEPT, RESERVED_SYMBOLS,
+                    OrderedSet, IDPZ3Error, DEF_SEMANTICS, Semantics)
 
 
 # Class Vocabulary  #######################################################
@@ -42,31 +46,45 @@ from .utils import (BOOL, INT, REAL, DATE, SYMBOL, OrderedSet, IDPZ3Error,
 def annotate(self, idp):
     self.idp = idp
 
+    # process Import and determine the constructors of CONCEPT
+    temp = {}  # contains the new self.declarations
+    for s in self.declarations:
+        if isinstance(s, Import):
+            other = self.idp.vocabularies[s.name]
+            for s1 in other.declarations:
+                if s1.name in temp:
+                    s.check(str(temp[s1.name]) == str(s1),
+                            f"Inconsistent declaration for {s1.name}")
+                temp[s1.name] = s1
+        else:
+            s.block = self
+            s.check(s.name not in temp or s.name in RESERVED_SYMBOLS,
+                    f"Duplicate declaration of {s.name}")
+            temp[s.name] = s
+    temp[CONCEPT].constructors=([Constructor(name=f"`{s}")
+                                 for s in [BOOL, INT, REAL, DATE, CONCEPT]]
+                               +[Constructor(name=f"`{s.name}")
+                                 for s in temp.values()
+                                 if s.name not in RESERVED_SYMBOLS
+                                 and type(s) in Declaration.__args__])
+    self.declarations = list(temp.values())
+
     # annotate declarations
     for s in self.declarations:
-        s.block = self
         s.annotate(self)  # updates self.symbol_decls
 
-    for constructor in self.symbol_decls[SYMBOL].constructors:
+    concepts = self.symbol_decls[CONCEPT]
+    for constructor in concepts.constructors:
         constructor.symbol = (Symbol(name=constructor.name[1:])
                                 .annotate(self, {}))
 
-    # populate .map of SYMBOL
-    for c in self.symbol_decls[SYMBOL].constructors:
+    # populate .map of CONCEPT
+    for c in concepts.constructors:
         assert not c.sorts
-        self.symbol_decls[SYMBOL].map[str(c)] = UnappliedSymbol.construct(c)
+        concepts.map[str(c)] = UnappliedSymbol.construct(c)
 Vocabulary.annotate = annotate
 
 
-# Class Extern  #######################################################
-
-def annotate(self, voc):
-    other = voc.idp.vocabularies[self.name]
-    #TODO merge while respecting order
-    voc.symbol_decls = {**other.symbol_decls, **voc.symbol_decls}
-Extern.annotate = annotate
-
-
 # Class TypeDeclaration  #######################################################
 
 def annotate(self, voc):
@@ -78,7 +96,7 @@ def annotate(self, voc):
     self.out.annotate(voc, {})
     for c in self.constructors:
         c.type = self.name
-        self.check(c.name not in voc.symbol_decls or self.name == SYMBOL,
+        self.check(c.name not in voc.symbol_decls or self.name == CONCEPT,
                     f"duplicate '{c.name}' constructor for '{self.name}' type")
         voc.symbol_decls[c.name] = c
     if self.interpretation:
@@ -97,11 +115,30 @@ def annotate(self, voc):
     for s in self.sorts:
         s.annotate(voc, {})
     self.out.annotate(voc, {})
-    self.type = self.out.decl.name
+    self.type = self.out.name
+
+    for s in chain(self.sorts, [self.out]):
+        self.check(s.name != CONCEPT or s == s, # use equality to check nested concepts
+                   f"`Concept` must be qualified with a type signature in {self}")
+    self.base_type = (None if self.out.name != BOOL or self.arity != 1 else
+                      self.sorts[0].decl.base_type)
     return self
 SymbolDeclaration.annotate = annotate
 
 
+# Class VarDeclaration  #######################################################
+
+def annotate(self, voc):
+    self.check(self.name not in voc.symbol_decls,
+                f"duplicate declaration in vocabulary: {self.name}")
+    self.check(self.name == self.name.rstrip(string.digits),
+                f"Variable {self.name} cannot be declared with a digital suffix.")
+    voc.symbol_decls[self.name] = self
+    self.subtype.annotate(voc, {})
+    return self
+VarDeclaration.annotate = annotate
+
+
 # Class Symbol  #######################################################
 
 def annotate(self, voc, q_vars):
@@ -113,7 +150,18 @@ def annotate(self, voc, q_vars):
 Symbol.annotate = annotate
 
 
-# Class Theory  #######################################################
+# Class Type  #######################################################
+
+def annotate(self, voc, q_vars={}):
+    Symbol.annotate(self, voc, q_vars)
+    if self.out:
+        self.ins = [s.annotate(voc, q_vars) for s in self.ins]
+        self.out = self.out.annotate(voc, q_vars)
+    return self
+Type.annotate = annotate
+
+
+# Class TheoryBlock  #######################################################
 
 def annotate(self, idp):
     self.check(self.vocab_name in idp.vocabularies,
@@ -124,18 +172,18 @@ def annotate(self, idp):
         i.annotate(self)
     self.voc.add_voc_to_block(self)
 
-    self.definitions = [e.annotate(self, self.voc, {}) for e in self.definitions]
+    self.definitions = [e.annotate(self.voc, {}) for e in self.definitions]
 
     self.constraints = OrderedSet([e.annotate(self.voc, {})
                                     for e in self.constraints])
-Theory.annotate = annotate
+TheoryBlock.annotate = annotate
 
 
 # Class Definition  #######################################################
 
-def annotate(self, theory, voc, q_vars):
+def annotate(self, voc, q_vars):
     self.rules = [r.annotate(voc, q_vars) for r in self.rules]
-    self.set_level_symbols()
+    self.set_level_symbols(voc)
 
     # create common variables, and rename vars in rule
     self.canonicals = {}
@@ -156,12 +204,15 @@ def annotate(self, theory, voc, q_vars):
     for decl, rules in self.canonicals.items():
         new_rule = copy(rules[0])
         exprs = [rule.body for rule in rules]
-        new_rule.body = ADisjunction.make('∨', exprs)
+        new_rule.body = OR(exprs)
         self.clarks[decl] = new_rule
     return self
 Definition.annotate = annotate
 
-def get_instantiables(self, for_explain=False):
+def get_instantiables(self,
+                      interpretations: Dict[str, SymbolInterpretation],
+                      extensions: Dict[str, Extension],
+                      for_explain=False):
     """ compute Definition.instantiables, with level-mapping if definition is inductive
 
     Uses implications instead of equivalence if `for_explain` is True
@@ -177,20 +228,20 @@ def get_instantiables(self, for_explain=False):
     result = {}
     for decl, rules in self.canonicals.items():
         rule = rules[0]
-        rule.is_whole_domain = all(s.decl.range  # not None nor []
-                               for s in rule.definiendum.decl.sorts)
+        rule.is_whole_domain = all(s.extension(interpretations, extensions)[0] is not None
+                                   for s in rule.definiendum.decl.sorts)
         if not rule.is_whole_domain:
             self.check(rule.definiendum.symbol.decl not in self.level_symbols,
                        f"Cannot have inductive definitions on infinite domain")
         else:
             if rule.out:
                 expr = AppliedSymbol.make(rule.definiendum.symbol,
-                                        rule.definiendum.sub_exprs[:-1])
+                                          rule.definiendum.sub_exprs[:-1])
                 expr.in_head = True
-                head = AComparison.make('=', [expr, rule.definiendum.sub_exprs[-1]])
+                head = EQUALS([expr, rule.definiendum.sub_exprs[-1]])
             else:
                 head = AppliedSymbol.make(rule.definiendum.symbol,
-                                        rule.definiendum.sub_exprs)
+                                          rule.definiendum.sub_exprs)
                 head.in_head = True
 
             inductive = (not rule.out and DEF_SEMANTICS != Semantics.COMPLETION
@@ -202,35 +253,29 @@ def get_instantiables(self, for_explain=False):
                 if not inductive:
                     bodies.append(r.body)
                     if for_explain and 1 < len(rules):  # not simplified -> no need to make copies
-                        out.append(ARImplication.make('⇐', [head, r.body],
-                                                      r.annotations))
+                        out.append(RIMPLIES([head, r.body], r.annotations))
                 else:
                     new = r.body.split_equivalences()
                     bodies.append(new)
                     if for_explain:
                         new = new.copy().add_level_mapping(rule.parent.level_symbols,
                                              rule.definiendum, False, False)
-                        out.append(ARImplication.make('⇐', [head, new],
-                                                      r.annotations))
+                        out.append(RIMPLIES([head, new], r.annotations))
 
-            all_bodies = ADisjunction.make('∨', bodies)
+            all_bodies = OR(bodies)
             if not inductive:
                 if out:  # already contains reverse implications
-                    out.append(ARImplication.make('⇒', [head, all_bodies],
-                                                  self.annotations))
+                    out.append(IMPLIES([head, all_bodies], self.annotations))
                 else:
-                    out = [AEquivalence.make('⇔', [head, all_bodies],
-                                             self.annotations)]
+                    out = [EQUIV([head, all_bodies], self.annotations)]
             else:
                 if not out:  # no reverse implication yet
                     new = all_bodies.copy().add_level_mapping(rule.parent.level_symbols,
                                              rule.definiendum, False, False)
-                    out = [ARImplication.make('⇐', [head.copy(), new],
-                                              self.annotations)]
+                    out = [RIMPLIES([head.copy(), new], self.annotations)]
                 all_bodies = all_bodies.copy().add_level_mapping(rule.parent.level_symbols,
                                         rule.definiendum, True, True)
-                out.append(AImplication.make('⇒', [head, all_bodies],
-                                             self.annotations))
+                out.append(IMPLIES([head, all_bodies], self.annotations))
             result[decl] = out
     return result
 Definition.get_instantiables = get_instantiables
@@ -265,12 +310,12 @@ def rename_args(self, new_vars):
         output: '!nv: f(nv) <- nv=args & body(args)'
     """
     self.check(len(self.definiendum.sub_exprs) == len(new_vars), "Internal error")
-    vars = [var.name for q in self.quantees for vars in q.vars for var in vars]
+    vars = {var.name : var for q in self.quantees for vars in q.vars for var in vars}
     for i in range(len(self.definiendum.sub_exprs)):
         arg, nv = self.definiendum.sub_exprs[i], list(new_vars.values())[i]
         if type(arg) == Variable \
         and arg.name in vars and arg.name not in new_vars:
-            vars.remove(arg.name)
+            del vars[arg.name]
             self.body = self.body.instantiate([arg], [nv])
             self.out = (self.out.instantiate([arg], [nv]) if self.out else
                         self.out)
@@ -278,10 +323,11 @@ def rename_args(self, new_vars):
                 self.definiendum.sub_exprs[j] = \
                     self.definiendum.sub_exprs[j].instantiate([arg], [nv])
         else:
-            eq = AComparison.make('=', [nv, arg])
-            self.body = AConjunction.make('∧', [eq, self.body])
+            eq = EQUALS([nv, arg])
+            self.body = AND([eq, self.body])
 
-    self.check(not vars, f"Too many variables in head of rule: {self}")
+    for v in vars.values():
+        self.body = EXISTS([Quantee.make(v, v.sort)], self.body)
 
     self.definiendum.sub_exprs = list(new_vars.values())
     self.quantees = [Quantee.make(v, v.sort) for v in new_vars.values()]
@@ -321,27 +367,34 @@ def annotate(self, block):
     voc = block.voc
     self.block = block
     self.symbol = Symbol(name=self.name).annotate(voc, {})
+    enumeration = self.enumeration  # shorthand
 
     # create constructors if it is a type enumeration
     self.is_type_enumeration = (type(self.symbol.decl) != SymbolDeclaration)
-    if self.is_type_enumeration and self.enumeration.constructors:
+    if self.is_type_enumeration and enumeration.constructors:
         # create Constructors before annotating the tuples
-        for c in self.enumeration.constructors:
+        for c in enumeration.constructors:
             c.type = self.name
             self.check(c.name not in voc.symbol_decls,
                     f"duplicate '{c.name}' constructor for '{self.name}' symbol")
             voc.symbol_decls[c.name] = c  #TODO risk of side-effects => use local decls ? issue #81
 
-    self.enumeration.annotate(voc)
+    enumeration.annotate(voc)
 
-    # predicate enumeration have FALSE default
-    if type(self.enumeration) != FunctionEnum and self.default is None:
-        self.default = FALSE
     self.check(self.is_type_enumeration
-                or all(s.name not in [INT, REAL, DATE]  # finite domain
+                or all(s.name not in [INT, REAL, DATE]  # finite domain #TODO
                         for s in self.symbol.decl.sorts)
                 or self.default is None,
         f"Can't use default value for '{self.name}' on infinite domain nor for type enumeration.")
+
+    self.check(not(self.symbol.decl.out.decl.base_type.name == BOOL
+                   and type(enumeration) == FunctionEnum),
+        f"Can't use function enumeration for predicates '{self.name}' (yet)")
+
+    # predicate enumeration have FALSE default
+    if type(enumeration) != FunctionEnum and self.default is None:
+        self.default = FALSE
+
     if self.default is not None:
         self.default = self.default.annotate(voc, {})
         self.check(self.default.value is not None,
@@ -352,8 +405,9 @@ SymbolInterpretation.annotate = annotate
 # Class Enumeration  #######################################################
 
 def annotate(self, voc):
-    for t in self.tuples:
-        t.annotate(voc)
+    if self.tuples:
+        for t in self.tuples:
+            t.annotate(voc)
 Enumeration.annotate = annotate
 
 
@@ -389,13 +443,13 @@ def annotate(self, voc):
         self.check(a.type in voc.symbol_decls,
                    f"Unknown type: {a.type}" )
         a.decl = SymbolDeclaration(annotations='', name=a.accessor,
-                                   sorts=[Symbol(name=self.type)],
-                                   out=Symbol(name=a.type))
+                                   sorts=[Type(name=self.type)],
+                                   out=Type(name=a.type))
         a.decl.annotate(voc)
     self.tester = SymbolDeclaration(annotations='',
                                     name=Symbol(name=f"is_{self.name}"),
-                                    sorts=[Symbol(name=self.type)],
-                                    out=Symbol(name=BOOL))
+                                    sorts=[Type(name=self.type)],
+                                    out=Type(name=BOOL))
     self.tester.annotate(voc)
 Constructor.annotate = annotate
 
@@ -433,9 +487,7 @@ def annotate(self, idp):
         open_types[name] = Symbol(name=type_name)
 
     for name, out in [
-        ('goal', Symbol(name=BOOL)),
         ('expand', Symbol(name=BOOL)),
-        ('relevant', Symbol(name=BOOL)),
         ('hide', Symbol(name=BOOL)),
         ('view', Symbol(name='_ViewType')),
         ('moveSymbols', Symbol(name=BOOL)),
@@ -444,23 +496,36 @@ def annotate(self, idp):
         ('optionalRelevance', Symbol(name=BOOL)),
         ('manualRelevance', Symbol(name=BOOL)),
         ('unit', open_types['unit']),
-        ('heading', open_types['heading'])
+        ('heading', open_types['heading']),
+        ('noOptimization', Symbol(name=BOOL))
     ]:
         symbol_decl = SymbolDeclaration(annotations='',
                                         name=Symbol(name=name),
-                                        sorts=[], out=out)
+                                        sorts=[out] if out else [], out=out)
         symbol_decl.annotate(self.voc)
 
-    # annotate constraints
+    # annotate constraints and interpretations
     for constraint in self.constraints:
         constraint.annotate(self.voc, {})
+    for i in self.interpretations.values():
+        i.annotate(self)
 Display.annotate = annotate
 
 
 # Class Expression  #######################################################
 
 def annotate(self, voc, q_vars):
-    " annotate tree after parsing "
+    """annotate tree after parsing
+
+    Resolve names and determine type as well as variables in the expression
+
+    Args:
+        voc (Vocabulary): the vocabulary
+        q_vars (Dict[str, Variable]): the quantifier variables that may appear in the expression
+
+    Returns:
+        Expression: an equivalent AST node, with updated type, .variables
+    """
     self.sub_exprs = [e.annotate(voc, q_vars) for e in self.sub_exprs]
     return self.annotate1()
 Expression.annotate = annotate
@@ -468,24 +533,24 @@ Expression.annotate = annotate
 
 def annotate1(self):
     " annotations that are common to __init__ and make() "
-    self.fresh_vars = set()
+    self.variables = set()
     if self.value is not None:
         pass
     if self.simpler is not None:
-        self.fresh_vars = self.simpler.fresh_vars
+        self.variables = self.simpler.variables
     else:
         for e in self.sub_exprs:
-            self.fresh_vars.update(e.fresh_vars)
+            self.variables.update(e.variables)
     return self
 Expression.annotate1 = annotate1
 
 
-# Class IfExpr  #######################################################
+# Class AIfExpr  #######################################################
 
 def annotate1(self):
-    self.type = self.sub_exprs[IfExpr.THEN].type
+    self.type = self.sub_exprs[AIfExpr.THEN].type
     return Expression.annotate1(self)
-IfExpr.annotate1 = annotate1
+AIfExpr.annotate1 = annotate1
 
 
 # Class AQuantification  #######################################################
@@ -497,10 +562,20 @@ def annotate(self, voc, q_vars):
         q.annotate(voc, q_vars)
         for vars in q.vars:
             for var in vars:
-                self.check(var.name not in voc.symbol_decls,
+                self.check(var.name not in voc.symbol_decls
+                           or type(voc.symbol_decls[var.name]) == VarDeclaration,
                     f"the quantified variable '{var.name}' cannot have"
                     f" the same name as another symbol")
                 var.sort = q.sub_exprs[0] if q.sub_exprs else None
+                var_decl = voc.symbol_decls.get(var.name.rstrip(string.digits), None)
+                if var_decl:
+                    subtype = var_decl.subtype
+                    self.check(var.sort is None
+                               or var.sort.name == subtype.name,
+                        f"Can't use declared {var.name} as a "
+                        f"{var.sort.name if var.sort else ''}")
+                    if var.sort is None:
+                        q.sub_exprs = [subtype.annotate(voc, {})]
                 q_v[var.name] = var
     self.sub_exprs = [e.annotate(voc, q_v) for e in self.sub_exprs]
     return self.annotate1()
@@ -511,10 +586,10 @@ def annotate1(self):
     for q in self.quantees:  # remove declared variables
         for vs in q.vars:
             for v in vs:
-                self.fresh_vars.discard(v.name)
+                self.variables.discard(v.name)
     for q in self.quantees:  # add variables in sort expression
         for sort in q.sub_exprs:
-            self.fresh_vars.update(sort.fresh_vars)
+            self.variables.update(sort.variables)
     return self
 AQuantification.annotate1 = annotate1
 
@@ -540,7 +615,7 @@ def annotate1(self):
 AImplication.annotate1 = annotate1
 
 
-# Class AImplication, AEquivalence  #######################################################
+# Class AEquivalence  #######################################################
 
 def annotate1(self):
     self.check(len(self.sub_exprs) == 2,
@@ -569,7 +644,7 @@ def annotate(self, voc, q_vars):
     out.type = BOOL
     # a≠b --> Not(a=b)
     if len(self.sub_exprs) == 2 and self.operator == ['≠']:
-        out = AUnary.make('¬', AComparison.make('=', self.sub_exprs))
+        out = NOT(EQUALS(self.sub_exprs))
     return out
 AComparison.annotate = annotate
 
@@ -588,14 +663,49 @@ AUnary.annotate1 = annotate1
 
 def annotate(self, voc, q_vars):
     self = AQuantification.annotate(self, voc, q_vars)
-    self.type = self.sub_exprs[AAggregate.OUT].type if self.out else INT
 
-    if not self.using_if:
-        self.sub_exprs = [IfExpr.make(if_f=self.sub_exprs[AAggregate.CONDITION],
-                then_f=Number(number='1') if self.out is None else
-                        self.sub_exprs[AAggregate.OUT],
-                else_f=Number(number='0'))]
-        self.using_if = True
+    if not self.annotated:
+        assert len(self.sub_exprs) == 1, "Internal error"
+        if self.aggtype == "#":
+            self.sub_exprs = [AIfExpr.make(self.sub_exprs[0],
+                                          Number(number='1'),
+                                          Number(number='0'))]
+            self.type = INT
+        else:
+            self.type = self.sub_exprs[0].type
+            if self.aggtype in ["min", "max"]:
+                # the `min` aggregate in `!y in T: min(lamda x in type: term(x,y))=0`
+                # is replaced by `_*(y)` with the following co-constraint:
+                #     !y in T: ( ?x in type: term(x) = _*(y)
+                #                !x in type: term(x) =< _*(y).
+                name = "_" + self.str
+                if name in voc.symbol_decls:
+                    symbol_decl = voc.symbol_decls[name]
+                    to_create = False
+                else:
+                    symbol_decl = SymbolDeclaration.make(
+                        "_"+self.str, # name `_ *`
+                        len(q_vars),  # arity
+                        [Type(name=v.sort.code) for v in q_vars.values()],
+                        Type(name=self.type)).annotate(voc)    # output_domain
+                    to_create = True
+                symbol = Symbol(name=symbol_decl.name)
+                applied = AppliedSymbol.make(symbol, q_vars.values())
+                applied = applied.annotate(voc, q_vars)
+
+                if to_create:
+                    coc1 = EXISTS(self.quantees,
+                                EQUALS([applied.copy(), self.sub_exprs[0]]))
+                    op = '≤' if self.aggtype == "min" else '≥'
+                    coc2 = FORALL(self.quantees.copy(),
+                                AComparison.make(op,
+                                        [applied.copy(), self.sub_exprs[0].copy()]))
+                    coc = AND([coc1, coc2])
+                    quantees = [Quantee.make(v, v.sort) for v in q_vars.values()]
+                    applied.co_constraint = FORALL(quantees, coc).annotate(voc, q_vars)
+                    applied.co_constraint.annotations['reading'] = f"Calculation of {self.code}"
+                return applied
+        self.annotated = True
     return self
 AAggregate.annotate = annotate
 AAggregate.annotate1 = AQuantification.annotate1
@@ -605,6 +715,9 @@ AAggregate.annotate1 = AQuantification.annotate1
 
 def annotate(self, voc, q_vars):
     self.symbol = self.symbol.annotate(voc, q_vars)
+    self.check((not self.symbol.decl or type(self.symbol.decl) != Constructor
+                or 0 < self.symbol.decl.arity),
+               f"Constructor `{self.symbol}` cannot be applied to argument(s)")
     self.sub_exprs = [e.annotate(voc, q_vars) for e in self.sub_exprs]
     if self.in_enumeration:
         self.in_enumeration.annotate(voc)
@@ -614,19 +727,19 @@ def annotate(self, voc, q_vars):
     if 'not' in self.is_enumerated:
         out = AppliedSymbol.make(out.symbol, out.sub_exprs,
                                  is_enumerated='is enumerated')
-        out = AUnary.make('¬', out)
+        out = NOT(out)
     elif 'not' in self.is_enumeration:
         out = AppliedSymbol.make(out.symbol, out.sub_exprs,
                                  is_enumeration='in',
                                  in_enumeration=out.in_enumeration)
-        out = AUnary.make('¬', out)
+        out = NOT(out)
     return out
 AppliedSymbol.annotate = annotate
 
 def annotate1(self):
     out = Expression.annotate1(self)
     out.symbol = out.symbol.annotate1()
-    out.fresh_vars.update(out.symbol.fresh_vars)
+    out.variables.update(out.symbol.variables)
     return out.simplify1()
 AppliedSymbol.annotate1 = annotate1
 
@@ -647,21 +760,31 @@ def annotate(self, voc, q_vars):
 Variable.annotate = annotate
 
 
+# Class Number  #######################################################
+
+def annotate(self, voc, q_vars):
+    self.decl = voc.symbol_decls[self.type]
+    return self
+Number.annotate = annotate
+
+
 # Class UnappliedSymbol  #######################################################
 
 def annotate(self, voc, q_vars):
+    if self.name in q_vars:  # ignore VarDeclaration
+        return q_vars[self.name]
     if self.name in voc.symbol_decls:
         self.decl = voc.symbol_decls[self.name]
-        self.fresh_vars = {}
+        self.variables = {}
         self.check(type(self.decl) == Constructor,
                    f"{self} should be applied to arguments (or prefixed with a back-tick)")
         return self
-    if self.name in q_vars:
-        return q_vars[self.name]
     # elif self.name in voc.symbol_decls:  # in symbol_decls
     #     out = AppliedSymbol.make(self.s, self.sub_exprs)
     #     return out.annotate(voc, q_vars)
     # If this code is reached, an undefined symbol was present.
+    if self.name.rstrip(string.digits) in q_vars:  # after considering it as a declared symbol
+        return self
     self.check(False, f"Symbol not in vocabulary: {self}")
 UnappliedSymbol.annotate = annotate
 
@@ -674,7 +797,7 @@ def annotate1(self):
     self.type = self.sub_exprs[0].type
     if self.annotations['reading']:
         self.sub_exprs[0].annotations = self.annotations
-    self.fresh_vars = self.sub_exprs[0].fresh_vars
+    self.variables = self.sub_exprs[0].variables
     return self
 Brackets.annotate1 = annotate1
 
