@@ -25,23 +25,25 @@ __all__ = ["Status", "Assignment", "Assignments"]
 
 from copy import copy
 from enum import Enum, auto
-from typing import Optional
+from typing import Dict, Optional, Tuple
 from z3 import BoolRef
 
-from .Expression import Expression, TRUE, FALSE, AUnary, AComparison
+from .Expression import Expression, TRUE, FALSE, NOT, EQUALS, AppliedSymbol
 from .utils import NEWL, BOOL
 
 
 class Status(Enum):
     """Describes how the value of a question was obtained"""
     UNKNOWN = auto()
-    GIVEN = auto()
-    ENV_UNIV = auto()
-    UNIVERSAL = auto()
-    ENV_CONSQ = auto()
-    CONSEQUENCE = auto()
-    EXPANDED = auto()
+    # fixed values:
     STRUCTURE = auto()
+    UNIVERSAL = auto()
+    CONSEQUENCE = auto()
+    ENV_CONSQ = auto()
+    # choices:
+    EXPANDED = auto()
+    DEFAULT = auto()
+    GIVEN = auto()
 
 
 class Assignment(object):
@@ -64,6 +66,8 @@ class Assignment(object):
 
         status (Status, optional): qualifies how the value was obtained
 
+        is_certainly_undefined (bool): True for functions applied to arguments certainly outside of its domain
+
         relevant (bool, optional): states whether the sentence is relevant
 
         symbol_decl (SymbolDeclaration): declaration of the symbol under which
@@ -71,15 +75,20 @@ class Assignment(object):
     """
     def __init__(self, sentence: Expression, value: Optional[Expression],
                  status: Optional[Status],
-                 relevant: Optional[bool] = False):
-        self.sentence = sentence
-        self.value = value
-        self.status = status
-        self.relevant = relevant
+                 relevant: Optional[bool] = True):
+        self.sentence: Expression = sentence
+        self.value: Optional[Expression] = value
+        self.status: Optional[Status] = status
+        self.is_certainly_undefined = False
+        self.relevant: Optional[bool] = relevant
 
-        # first symbol in the sentence, preferably not starting with '_'
-        self.symbol_decl, default = None, None
-        self.symbols = sentence.collect_symbols(co_constraints=False).values()
+        # First symbol in the sentence, preferably not starting with '_':
+        # if no public symbol (not starting with '_') is found, the first
+        # private one is used.
+        self.symbol_decl: "SymbolDeclaration" = None
+        default = None
+        self.symbols: Dict[str, "SymbolDeclaration"] = \
+            sentence.collect_symbols(co_constraints=False).values()
         for d in self.symbols:
             if not d.private:
                 if d.block:  # ignore accessors and testers
@@ -90,12 +99,13 @@ class Assignment(object):
         if not self.symbol_decl:  # use the '_' symbol (to allow relevance computation)
             self.symbol_decl = default
 
-    def copy(self):
+    def copy(self, shallow: Optional[bool] =False) -> "Assignment":
         out = copy(self)
-        out.sentence = out.sentence.copy()
+        if not shallow:
+            out.sentence = out.sentence.copy()
         return out
 
-    def __str__(self):
+    def __str__(self) -> str:
         pre, post = '', ''
         if self.value is None:
             pre = "? "
@@ -107,10 +117,10 @@ class Assignment(object):
             post = f" -> {str(self.value)}"
         return f"{pre}{self.sentence.annotations['reading']}{post}"
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return self.__str__()
 
-    def __log__(self):
+    def __log__(self) -> Optional[Expression]:
         return self.value
 
     def same_as(self, other:"Assignment") -> bool:
@@ -130,17 +140,17 @@ class Assignment(object):
     def to_json(self) -> str:  # for GUI
         return str(self)
 
-    def formula(self):
+    def formula(self) -> Expression:
         if self.value is None:
             raise Exception("can't translate unknown value")
         if self.sentence.type == BOOL:
             out = self.sentence if self.value.same_as(TRUE) else \
-                AUnary.make('¬', self.sentence)
+                NOT(self.sentence)
         else:
-            out = AComparison.make('=', [self.sentence, self.value])
+            out = EQUALS([self.sentence, self.value])
         return out
 
-    def negate(self):
+    def negate(self) -> "Assignment":
         """returns an Assignment for the same sentence, but an opposite truth value.
 
         Raises:
@@ -153,10 +163,13 @@ class Assignment(object):
         value = FALSE if self.value.same_as(TRUE) else TRUE
         return Assignment(self.sentence, value, self.status, self.relevant)
 
-    def translate(self) -> BoolRef:
-        return self.formula().translate()
+    def translate(self, problem: "Theory") -> BoolRef:
+        return self.formula().translate(problem)
 
-    def as_set_condition(self):
+    def as_set_condition(self
+             ) -> Tuple[Optional[AppliedSymbol],
+                        Optional[bool],
+                        Optional["Enumeration"]]:
         """returns an equivalent set condition, or None
 
         Returns:
@@ -167,7 +180,7 @@ class Assignment(object):
             return (x, y if self.value.same_as(TRUE) else not y, z)
         return (None, None, None)
 
-    def unset(self):
+    def unset(self) -> None:
         """ Unsets the value of an assignment.
 
         Returns:
@@ -180,44 +193,77 @@ class Assignments(dict):
     """Contains a set of Assignment"""
     def __init__(self, *arg, **kw):
         super(Assignments, self).__init__(*arg, **kw)
-        self.symbols = {}  # { decl.name: decl }
+        self.symbols: Dict[str, "SymbolDeclaration"] = {}
         for a in self.values():
             if a.symbol_decl:
                 self.symbols[a.symbol_decl.name] = a.symbol_decl
 
-    def copy(self):
-        return Assignments({k: v.copy() for k, v in self.items()})
+    def copy(self, shallow: bool = False) -> "Assignments":
+        return Assignments({k: v.copy(shallow) for k, v in self.items()})
 
-    def extend(self, more):
+    def extend(self, more: "Assignments") -> None:
         for v in more.values():
             self.assert_(v.sentence, v.value, v.status, v.relevant)
 
-    def assert__(self, sentence: Expression,
-                value: Optional[Expression],
-                status: Optional[Status],
-                relevant: Optional[bool]):
+    def assert__(self,
+                 sentence: Expression,
+                 value: Optional[Expression],
+                 status: Optional[Status]
+                ) -> Assignment:
+
         if sentence.code in self:
-            out = self[sentence.code]
-            out.value = value
-            if not (out.status == Status.ENV_CONSQ and status == Status.CONSEQUENCE):
-                # do not change an env consequence to a decision consequence
-                out.status = status
-            if relevant is not None:
-                out.relevant = relevant
+            out = self[sentence.code].copy(shallow=True)
+            if out.status in [Status.GIVEN, Status.EXPANDED, Status.DEFAULT]\
+                    and status in [Status.CONSEQUENCE, Status.ENV_CONSQ, Status.UNIVERSAL]:
+                assert out.value.same_as(value), \
+                        "System should not override given choices with different consequences, please report this bug."
+            else:
+                if not (out.status == Status.ENV_CONSQ and status == Status.CONSEQUENCE):
+                    # do not change an env consequence to a decision consequence
+                    out.value = value
+                    out.status = status
         else:
-            out = Assignment(sentence, value, status, relevant)
+            out = Assignment(sentence, value, status)
         if out.symbol_decl:  # ignore comparisons of constructors
             self[sentence.code] = out
             self.symbols[out.symbol_decl.name] = out.symbol_decl
         return out
 
-    def __str__(self):
+    def __str__(self) -> str:
+        """ Print the assignments in the same format as a model.
+
+        Most symbols are printed as `name := {(val1, ..., val} -> valx, ...}.`
+        with the exception of nullary symbols, which are printed as
+        `name := value`.
+        """
         out = {}
+        nullary = set()
         for a in self.values():
-            if a.value is not None and not a.sentence.is_reified():
-                c = ",".join(str(e) for e in a.sentence.sub_exprs)
+            if (a.value is not None and not a.sentence.is_reified()):
+                # make sure we have an entry for `a` in `out`
+                if a.value == FALSE and a.symbol_decl.arity != 0:
+                    if a.symbol_decl.name not in out:
+                        out[a.symbol_decl.name] = []
+                    continue
+
+                c = ", ".join(str(e) for e in a.sentence.sub_exprs)
                 c = f"({c})" if 1 < len(a.sentence.sub_exprs) else c
-                c = f"{c}->{str(a.value)}"
+                if a.symbol_decl.arity == 0:
+                    # Symbol is a proposition or constant.
+                    c = f"{c}"
+                    nullary.add(a.symbol_decl.name)
+                if a.value == TRUE and a.symbol_decl.arity > 0:
+                    # Symbol is a predicate.
+                    c = f"{c}"
+                else:
+                    # Symbol is a function.
+                    c = f"{c} -> {str(a.value)}"
                 out[a.symbol_decl.name] = out.get(a.symbol_decl.name, []) + [c]
-        return NEWL.join(f"{k}:={{{ '; '.join(s for s in a) }}}"
-                         for k, a in out.items())
+
+        model_str = ""
+        for k, a in out.items():
+            if k in nullary:  # Exception 1.
+                model_str += f"{k} :={a[0][3:]}.{NEWL}"
+            else:
+                model_str += f"{k} := {{{ ', '.join(s for s in a) }}}.{NEWL}"
+        return model_str
