@@ -30,12 +30,13 @@ __all__ = ["ASTNode", "Expression", "Constructor", "AIfExpr", "Quantee", "AQuant
 
 import copy
 from collections import ChainMap
-from datetime import date
+from datetime import datetime, date
+from dateutil.relativedelta import *
 from fractions import Fraction
 from re import findall
 from sys import intern
 from textx import get_location
-from typing import Optional, List, Tuple, Dict, Set, Any
+from typing import Optional, List, Tuple, Dict, Set, Any, Callable
 
 from .utils import unquote, OrderedSet, BOOL, INT, REAL, DATE, CONCEPT, RESERVED_SYMBOLS, \
     IDPZ3Error, DEF_SEMANTICS, Semantics
@@ -288,6 +289,8 @@ class Expression(ASTNode):
     def same_as(self, other):
         if self.str == other.str:
             return True
+        if self.__class__.__name__ == "Number" and other.__class__.__name__ == "Number":
+            return float(self.py_value) == float(other.py_value)
         if self.value is not None and self.value is not self:
             return self.value  .same_as(other)
         if self.simpler is not None:
@@ -506,6 +509,8 @@ class Expression(ASTNode):
                                    for e in self.sub_exprs))
                     .annotate1())  # update .variables
 
+Extension = Tuple[Optional[List[List[Expression]]],  # None if the extension is infinite (e.g., Int)
+                  Optional[Callable]]  # None if filtering is not required
 
 class Symbol(Expression):
     """Represents a Symbol.  Handles synonyms.
@@ -531,8 +536,11 @@ class Symbol(Expression):
     def __repr__(self):
         return str(self)
 
-    def has_element(self, term: Expression) -> Expression:
-        """Returns an expression that says whether `term` is in the type denoted by `self`.
+    def has_element(self, term: Expression,
+                    interpretations: Dict[str, "SymbolInterpretation"],
+                    extensions: Dict[str, Extension]
+                    ) -> Expression:
+        """Returns an expression that says whether `term` is in the type/predicate denoted by `self`.
 
         Args:
             term (Expression): the argument to be checked
@@ -540,10 +548,11 @@ class Symbol(Expression):
         Returns:
             Expression: whether `term` is in the type denoted by `self`.
         """
-        return self.decl.check_bounds(term)
+        self.check(self.decl.out.name == BOOL, "internal error")
+        return self.decl.contains_element(term, interpretations, extensions)
 
 
-class Subtype(Symbol):
+class Type(Symbol):
     """ASTNode representing `aType` or `Concept[aSignature]`, e.g., `Concept[T*T->Bool]`
 
     Inherits from Symbol
@@ -574,11 +583,16 @@ class Subtype(Symbol):
                     len(self.ins) == len(other.ins) and
                     all(s==o for s, o in zip(self.ins, other.ins)))))
 
-    def range():
+    def extension(self, interpretations: Dict[str, "SymbolInterpretation"],
+                  extensions: Dict[str, Extension]):
         pass  # monkey-patched
 
-    def has_element(self, term: Expression) -> Expression:
-        """Returns an Expression that says whether `term` is in the type denoted by `self`.
+    def has_element(self,
+                    term: Expression,
+                    interpretations: Dict[str, "SymbolInterpretation"],
+                    extensions: Dict[str, Extension]
+                    ) -> Expression:
+        """Returns an Expression that says whether `term` is in the type/predicate denoted by `self`.
 
         Args:
             term (Expression): the argument to be checked
@@ -587,10 +601,12 @@ class Subtype(Symbol):
             Expression: whether `term` `term` is in the type denoted by `self`.
         """
         if self.name == CONCEPT:
-            comparisons = [EQUALS([term, c]) for c in self.range()]
+            comparisons = [EQUALS([term, c[0]])
+                           for c in self.extension(interpretations, extensions)[0]]
             return OR(comparisons)
         else:
-            return self.decl.check_bounds(term)
+            self.check(self.decl.out.name == BOOL, "internal error")
+            return self.decl.contains_element(term, interpretations, extensions)
 
 
 class AIfExpr(Expression):
@@ -628,7 +644,7 @@ class Quantee(Expression):
     Attributes:
         vars (List[List[Variable]]): the (tuples of) variables being quantified
 
-        subtype (Subtype, Optional): a literal Subtype to quantify over, e.g., `Color` or `Concept[Color->Bool]`.
+        subtype (Type, Optional): a literal Type to quantify over, e.g., `Color` or `Concept[Color->Bool]`.
 
         sort (SymbolExpr, Optional): a dereferencing expression, e.g.,. `$(i)`.
 
@@ -1213,7 +1229,7 @@ class Variable(Expression):
         self.name = kwargs.pop('name')
         sort = kwargs.pop('sort') if 'sort' in kwargs else None
         self.sort = sort
-        assert sort is None or isinstance(sort, Subtype) or isinstance(sort, Symbol)
+        assert sort is None or isinstance(sort, Type) or isinstance(sort, Symbol)
 
         super().__init__()
 
@@ -1245,12 +1261,8 @@ class Number(Expression):
             self.py_value = Fraction(self.number)
             self.type = REAL
         elif '.' in self.number:
-            v = (self.number if not self.number.endswith('?') else
-                 self.number[:-1])
-            if "e" in v:
-                self.py_value = float(eval(v))
-            else:
-                self.py_value = Fraction(v)
+            self.py_value = Fraction(self.number if not self.number.endswith('?') else
+                                     self.number[:-1])
             self.type = REAL
         else:
             self.py_value = int(self.number)
@@ -1275,8 +1287,15 @@ class Date(Expression):
 
     def __init__(self, **kwargs):
         self.iso = kwargs.pop('iso')
-        self.date = (date.today() if self.iso == '#TODAY' else
+
+        dt = (date.today() if self.iso == '#TODAY' else
                      date.fromisoformat(self.iso[1:]))
+        if 'y' in kwargs:
+            y = int(kwargs.pop('y'))
+            m = int(kwargs.pop('m'))
+            d = int(kwargs.pop('d'))
+            dt = dt + relativedelta(years=y, months=m, days=d)
+        self.date = dt
 
         super().__init__()
 
@@ -1284,9 +1303,13 @@ class Date(Expression):
         self.variables = set()
         self.value = self
 
-        self.py_value = self.date.toordinal()
+        self.py_value = int(self.date.toordinal())
         self.type = DATE
 
+    @classmethod
+    def make(cls, value):
+        return cls(iso=f"#{date.fromordinal(value).isoformat()}")
+
     def __str__(self): return f"#{self.date.isoformat()}"
 
     def is_reified(self): return False
