@@ -37,7 +37,7 @@ from .Assignments import Status as S, Assignments
 from .Expression import (Expression, AQuantification,
                     ADisjunction, AConjunction, AppliedSymbol,
                     AComparison, AUnary, Brackets, TRUE, FALSE)
-from .Parse import str_to_IDP
+from .Parse import str_to_IDP, TypeDeclaration
 from .Theory import Theory
 from .utils import OrderedSet, IDPZ3Error, NOT_SATISFIABLE
 
@@ -183,7 +183,7 @@ def _set_consequences_get_changed_choices(self):
     # clear consequences, as these might not be cleared by add_given when
     # running via CLI
     for a in self.assignments.values():
-        if a.status in [S.CONSEQUENCE, S.ENV_CONSQ]:
+        if a.status in [S.CONSEQUENCE]:
             self.assignments.assert__(a.sentence, None, S.UNKNOWN)
 
     removed_choices = {a.sentence.code: a for a in self.previous_assignments.values()
@@ -223,7 +223,8 @@ def _directional_todo(self, removed_choices={}, added_choices=[]):
         for a in removed_choices.values():
             todo[a.sentence.code] = a.sentence
         for a in self.previous_assignments.values():
-            if a.status in [S.CONSEQUENCE, S.ENV_CONSQ]:
+            if (a.status in [S.CONSEQUENCE, S.ENV_CONSQ]
+                or a.is_certainly_undefined):
                 todo[a.sentence.code] = a.sentence
 
     if added_choices:
@@ -293,7 +294,6 @@ def _batch_propagate(self, tag=S.CONSEQUENCE):
         yield "No more consequences."
 Theory._batch_propagate = _batch_propagate
 
-
 def _propagate_inner(self, tag, solver, todo):
     for q in todo.values():
         solver.add(q.reified(self) == q.translate(self))
@@ -304,18 +304,24 @@ def _propagate_inner(self, tag, solver, todo):
     if res1 == sat:
         model = solver.model()
         valqs = [(model.eval(q.reified(self)), q) for q in todo.values()]
-        for val1, q in valqs:
+        while valqs:
+            (val1, q) = valqs.pop()
             if str(val1) == str(q.reified(self)):
                 continue  # irrelevant
-            solver.push()
-            solver.add(Not(q.reified(self) == val1))
-            res2 = solver.check()
-            solver.pop()
 
-            assert res2 != unknown, "Incorrect solver behavior"
-            if res2 == unsat:
-                val = str_to_IDP(q, str(val1))
-                yield self.assignments.assert__(q, val, tag)
+            is_certainly_undefined = self._is_undefined(solver, q)
+            if q.code in self.assignments:
+                self.assignments[q.code].is_certainly_undefined = is_certainly_undefined
+            if not is_certainly_undefined:
+                solver.push()
+                solver.add(Not(q.reified(self) == val1))
+                res2 = solver.check()
+                solver.pop()
+
+                assert res2 != unknown, "Incorrect solver behavior"
+                if res2 == unsat:
+                    val = str_to_IDP(q, str(val1))
+                    yield self.assignments.assert__(q, val, tag)
 
         yield "No more consequences."
     elif res1 == unsat:
@@ -323,7 +329,6 @@ def _propagate_inner(self, tag, solver, todo):
         yield str(solver.sexpr())
     else:
         assert False, "Incorrect solver behavior"
-
 Theory._propagate_inner = _propagate_inner
 
 
