@@ -28,14 +28,14 @@ import sys
 from typing import List
 
 from .Expression import (
-    Constructor, Expression, AIfExpr, AQuantification, Quantee, Subtype,
+    Constructor, Expression, AIfExpr, AQuantification, Quantee, Type,
     Operator, AEquivalence, AImplication, ADisjunction,
     AConjunction, AComparison, ASumMinus, AMultDiv, APower,
     AUnary, AAggregate, SymbolExpr, AppliedSymbol, UnappliedSymbol, Variable,
     Number, Date, Brackets, TRUE, FALSE, NOT, AND, OR)
 from .Parse import Symbol, Enumeration, Tuple, TypeDeclaration
 from .Assignments import Status as S, Assignment
-from .utils import BOOL, INT, CONCEPT, ABS, RESERVED_SYMBOLS
+from .utils import BOOL, INT, DATE, CONCEPT, ABS, RESERVED_SYMBOLS
 
 
 # class Expression  ###########################################################
@@ -59,7 +59,7 @@ def _change(self, sub_exprs=None, ops=None, value=None, simpler=None,
             self.simpler = simpler
     assert (self.value is None
                or type(self.value) in [AppliedSymbol, UnappliedSymbol, Symbol,
-                                       Number, Date, Subtype]), \
+                                       Number, Date, Type]), \
             f"Incorrect value in _change: {self.value}"
 
     # reset derived attributes
@@ -264,7 +264,8 @@ def update_arith(self, family, operands):
                 out //= e.py_value
             else:
                 out = function(out, e.py_value)
-        value = Number(number=str(out))
+        value = (Number(number=str(out)) if operands[0].type != DATE else
+                 Date.make(out))
         return self._change(value=value, sub_exprs=operands)
     return self._change(sub_exprs=operands)
 
