@@ -1,3 +1,4 @@
+#
 # Copyright 2019 Ingmar Dasseville, Pierre Carbonnelle
 #
 # This file is part of Interactive_Consultant.
@@ -31,13 +32,12 @@ from z3 import (Context, BoolRef, ExprRef, Solver, sat, unsat, Optimize, Not, An
 
 from .Assignments import Status as S, Assignment, Assignments
 from .Expression import (TRUE, Expression, FALSE, AppliedSymbol,
-                         EQUALS, NOT)
-from .Parse import (TypeDeclaration, Type, SymbolDeclaration, Symbol,
+                         EQUALS, NOT, Extension)
+from .Parse import (TypeDeclaration, Declaration, SymbolDeclaration, Symbol,
                     TheoryBlock, Structure, Definition, str_to_IDP, SymbolInterpretation)
 from .Simplify import join_set_conditions
 from .utils import (OrderedSet, NEWL, BOOL, INT, REAL, DATE,
                     RESERVED_SYMBOLS, CONCEPT, GOAL_SYMBOL, RELEVANT, NOT_SATISFIABLE)
-from .Idp_to_Z3 import get_symbols_z
 
 
 class Propagation(Enum):
@@ -60,7 +60,7 @@ class Theory(object):
         extended (Bool): True when the truth value of inequalities
             and quantified formula is of interest (e.g. in the Interactive Consultant)
 
-        declarations (Dict[str, Type]): the list of type and symbol declarations
+        declarations (Dict[str, Declaration]): the list of type and symbol declarations
 
         constraints (OrderedSet): a set of assertions.
 
@@ -69,14 +69,15 @@ class Theory(object):
         interpretations (Dict[string, SymbolInterpretation]):
             A mapping of enumerated symbols to their interpretation.
 
-        def_constraints (Dict[SymbolDeclaration, Definition], list[Expression]):
+        extensions (Dict[string, Extension]):
+            Extension of types and predicates
+
+        def_constraints (Dict[SymbolDeclaration, Definition], List[Expression]):
             A mapping of defined symbol to the whole-domain constraints
             equivalent to its definition.
 
         _constraintz (List(ExprRef), Optional): a list of assertions, co_constraints and definitions in Z3 form
 
-        _symbols (Set[str]): set of symbol name occurring in self._constraintz
-
         _formula (ExprRef, optional): the Z3 formula that represents
             the problem (assertions, co_constraints, definitions and assignments).
 
@@ -129,12 +130,13 @@ class Theory(object):
 
         self.extended: Optional[bool] = extended
 
-        self.declarations: Dict[str, Type] = {}
+        self.declarations: Dict[str, Declaration] = {}
         self.definitions: List[Definition] = []
         self.constraints: OrderedSet = OrderedSet()
         self.assignments: Assignments = Assignments()
         self.def_constraints: Dict[Tuple[SymbolDeclaration, Definition], List[Expression]] = {}
-        self.interpretations: Dict[str, SymbolInterpretation] = {}
+        self.interpretations: Dict[str, SymbolInterpretation] = {}  # interpretations given by user
+        self.extensions: Dict[str, Extension] = {}  # computed extension of types and predicates
         self.name: str = ''
 
         self._contraintz: Optional[List[BoolRef]] = None
@@ -166,8 +168,9 @@ class Theory(object):
             assignment_forms = [a.formula().translate(self)
                                 for a in self.assignments.values()
                                 if a.value is not None
-                                and a.status in [S.STRUCTURE, S.UNIVERSAL]
-                                and a.symbol_decl.name in self._symbols]
+                                and (a.status == S.UNIVERSAL
+                                     or (a.status == S.STRUCTURE
+                                         and a.symbol_decl.needs_interpretation))]
             self._slvr.add(assignment_forms)
         return self._slvr
 
@@ -180,8 +183,9 @@ class Theory(object):
             assignment_forms = [a.formula().translate(self)
                                 for a in self.assignments.values()
                                 if a.value is not None
-                                and a.status in [S.STRUCTURE, S.UNIVERSAL]
-                                and a.symbol_decl.name in self._symbols]
+                                and (a.status == S.UNIVERSAL
+                                     or (a.status == S.STRUCTURE
+                                         and a.symbol_decl.needs_interpretation))]
             self._optmz.add(assignment_forms)
         return self._optmz
 
@@ -194,13 +198,13 @@ class Theory(object):
             # get expanded def_constraints
             def_constraints = {}
             for defin in self.definitions:
-                instantiables = defin.get_instantiables(for_explain=True)
+                instantiables = defin.get_instantiables(self.interpretations, self.extensions, for_explain=True)
                 defin.add_def_constraints(instantiables, self, def_constraints)
 
-            todo = chain(self.constraints, chain(*def_constraints.values()))
-            for constraint in todo:
+            for constraint in chain([c.interpret(self) for c in self.constraints],
+                                    chain(*def_constraints.values())):
                 p = constraint.reified(self)
-                self.expl_reifs[p] = (constraint.original.interpret(self).translate(self), constraint)
+                self.expl_reifs[p] = (constraint.translate(self), constraint)
                 self._reif.add(Implies(p, self.expl_reifs[p][0]))
 
         return self._reif
@@ -267,8 +271,11 @@ class Theory(object):
         # apply the enumerations and definitions
 
         self.assignments = Assignments()
+        self.extensions = {}  # reset the cache
 
         for decl in self.declarations.values():
+            if type(decl) == SymbolDeclaration:  # reset it
+                decl.needs_interpretation = False
             decl.interpret(self)
 
         # remove RELEVANT constraints
@@ -277,21 +284,18 @@ class Theory(object):
                    and v.decl is not None
                    and v.decl.name == RELEVANT)])
 
-        # process enumerations, including GOAL_SYMBOL
-        for symbol_interpretation in self.interpretations.values():
-            if not symbol_interpretation.is_type_enumeration:
-                symbol_interpretation.interpret(self)
-            if symbol_interpretation.name == GOAL_SYMBOL:
-                # expand goal_symbol
-                for t in symbol_interpretation.enumeration.tuples:
-                    symbol = t.args[0]
-                    decl = self.declarations[symbol.name[1:]]
-                    assert decl.instances, f"goal {decl.name} must be instantiable."
-                    relevant = Symbol(name=RELEVANT)
-                    relevant.decl = self.declarations[RELEVANT]
-                    for i in decl.instances.values():
-                        constraint = AppliedSymbol.make(relevant, [i])
-                        self.constraints.append(constraint)
+        # expand goal_symbol
+        symbol_interpretation = self.interpretations.get(GOAL_SYMBOL, None)
+        if symbol_interpretation:
+            for t in symbol_interpretation.enumeration.tuples:
+                symbol = t.args[0]
+                decl = self.declarations[symbol.name[1:]]
+                assert decl.instances, f"goal {decl.name} must be instantiable."
+                relevant = Symbol(name=RELEVANT)
+                relevant.decl = self.declarations[RELEVANT]
+                for i in decl.instances.values():
+                    constraint = AppliedSymbol.make(relevant, [i])
+                    self.constraints.append(constraint)
 
         # expand whole-domain definitions
         for defin in self.definitions:
@@ -309,6 +313,9 @@ class Theory(object):
         for es in self.def_constraints.values():
             for e in es:
                 e.co_constraints(self.co_constraints)
+        for c in self.co_constraints.values():
+            c.interpret(self)
+
         for s in list(questions.values()):
             if s.code not in self.assignments:
                 self.assignments.assert__(s, None, S.UNKNOWN)
@@ -401,9 +408,6 @@ class Theory(object):
             self._constraintz += [s.translate(self)
                             for s in chain(*self.def_constraints.values())]
 
-            self._symbols = set()
-            for c in self._constraintz:
-                get_symbols_z(c, self._symbols)
         return self._constraintz
 
     def formula(self) -> BoolRef:
@@ -418,7 +422,7 @@ class Theory(object):
                 # occurring in the (potentially simplified) z3 constraints
                 all = ([a.formula().translate(self)
                         for a in self.assignments.values()
-                        if a.symbol_decl.name in self._symbols
+                        if (a.status != S.STRUCTURE or a.symbol_decl.needs_interpretation)
                         and a.value is not None
                         and a.status not in [S.CONSEQUENCE, S.ENV_CONSQ]]
                         + self.constraintz())
@@ -438,6 +442,41 @@ class Theory(object):
         s.add(self.formula())
         return s.sexpr()
 
+
+    def _is_defined(self, model, q):
+        # determine if the expression is defined
+        defined = True
+        if type(q) == AppliedSymbol:
+            if any(type(T.decl) != TypeDeclaration for T in q.decl.sorts):
+                in_domain = q.decl.has_in_domain(q.sub_exprs, self.interpretations, self.extensions)
+                if in_domain.same_as(FALSE):
+                    defined = False
+                elif in_domain.same_as(TRUE):
+                    defined = True
+                else:
+                    defined = model.eval(in_domain.translate(self))
+                    if str(defined) == str(in_domain):
+                        defined = True  #TODO dubious. Why not False ?
+        return defined
+
+    def _is_undefined(self, solver, q):
+        # determine if the expression is certainly undefined
+        result = False
+        if type(q) == AppliedSymbol:
+            if any(type(T.decl) != TypeDeclaration for T in q.decl.sorts):
+                in_domain = q.decl.has_in_domain(q.sub_exprs, self.interpretations, self.extensions)
+                if in_domain.same_as(FALSE):
+                    result = True
+                elif in_domain.same_as(TRUE):
+                    result = False
+                else:
+                    solver.push()
+                    solver.add(in_domain.translate(self))
+                    res = solver.check()
+                    solver.pop()
+                    result = res == unsat
+        return result
+
     def _from_model(self,
                     solver: Solver,
                     todo: List[Expression],
@@ -449,20 +488,27 @@ class Theory(object):
         ass = copy(self.assignments)
         model = solver.model()
         for q in todo:
-            assert (self.extended or not q.is_reified(),
-                    "Reified atom should only appear in case of extended theories")
-            if complete or q.is_reified():
-                val1 = model.eval(q.reified(self), model_completion=complete)
+            assert self.extended or not q.is_reified(), \
+                    "Reified atom should only appear in case of extended theories"
+
+            a = ass[q.code]
+            if not self._is_defined(model, q):
+                a.value, a.tag, a.relevant = None, S.UNKNOWN, False
             else:
-                val1 = model.eval(q.translate(self), model_completion=complete)
-            val = str_to_IDP(q, str(val1))
-            if val is not None:
-                if q.is_assignment() and val == FALSE:
-                    tag = (S.ENV_CONSQ if q.sub_exprs[0].decl.block.name == 'environment'
-                           else S.CONSEQUENCE)
+                if complete or q.is_reified():
+                    val1 = model.eval(q.reified(self), model_completion=complete)
                 else:
-                    tag = S.EXPANDED
-                ass.assert__(q, val, tag)
+                    val1 = model.eval(q.translate(self), model_completion=complete)
+                val = str_to_IDP(q, str(val1))
+                if val is not None:
+                    if q.is_assignment() and val == FALSE:  # consequence of the TRUE assignment
+                        tag = (S.ENV_CONSQ if q.sub_exprs[0].decl.block.name == 'environment'
+                            else S.CONSEQUENCE)
+                    else:
+                        tag = S.EXPANDED
+                    ass.assert__(q, val, tag)
+                else:
+                    a.value, a.tag, a.relevant = None, S.UNKNOWN, False
         return ass
 
     def _add_assignment(self, solver: Solver) -> None:
@@ -516,13 +562,13 @@ class Theory(object):
 
     def expand(self,
                max: int = 10,
-               timeout: int = 10,
+               timeout_seconds: int = 10,
                complete: bool = False
                ) -> Iterator[Union[Assignments, str]]:
         """Generates a list of models of the theory that are expansion of the known assignments.
 
         The result is limited to ``max`` models (10 by default), or unlimited if ``max`` is 0.
-        The search for new models is stopped when processing exceeds ``timeout`` (in seconds) (unless it is 0).
+        The search for new models is stopped when processing exceeds ``timeout_seconds`` (in seconds) (unless it is 0).
         The models can be asked to be complete or partial (i.e., in which "don't care" terms are not specified).
 
         The string message can be one of the following:
@@ -531,13 +577,13 @@ class Theory(object):
 
         - ``More models may be available.  Change the max argument to see them.``
 
-        - ``More models may be available.  Change the timeout argument to see them.``
+        - ``More models may be available.  Change the timeout_seconds argument to see them.``
 
-        - ``More models may be available.  Change the max and timeout arguments to see them.``
+        - ``More models may be available.  Change the max and timeout_seconds arguments to see them.``
 
         Args:
             max (int, optional): maximum number of models. Defaults to 10.
-            timeout (int, optional): timeout in seconds.  Defaults to 10.
+            timeout_seconds (int, optional): timeout_seconds seconds.  Defaults to 10.
             complete (bool, optional): ``True`` for complete models. Defaults to False.
 
         Yields:
@@ -568,7 +614,7 @@ class Theory(object):
         count, ass = 0, {}
         start = time.process_time()
         while ((max <= 0 or count < max) and
-               (timeout <= 0 or time.process_time()-start < timeout)):
+               (timeout_seconds <= 0 or time.process_time()-start < timeout_seconds)):
             # exclude ass
             different = []
             for a in ass.values():
@@ -590,12 +636,12 @@ class Theory(object):
         solver.pop()
 
         maxed = (0 < max <= count)
-        timeouted = (0 < timeout <= time.process_time()-start)
-        # if interrupted by the timeout
+        timeouted = (0 < timeout_seconds <= time.process_time()-start)
+        # if interrupted by the timeout_seconds
         if maxed or timeouted:
-            param = ("max and timeout arguments" if maxed and timeouted else
+            param = ("max and timeout_seconds arguments" if maxed and timeouted else
                      "max argument" if maxed else
-                     "timeout argument")
+                     "timeout_seconds argument")
             yield f"{NEWL}More models may be available.  Change the {param} to see them."
         elif 0 < count:
             yield f"{NEWL}No more models."
@@ -606,8 +652,11 @@ class Theory(object):
                  term: str,
                  minimize: bool = True
                  ) -> "Theory":
-        """Updates the Theory so that the value of term in the ``assignments`` property
-        is the optimal value that is compatible with the Theory.
+        """Updates the value of `term` in the ``assignments`` property of `self`
+        to the optimal value that is compatible with the theory.
+
+        Chain it with a call to `expand` to obtain a model,
+        or to `propagate` to propagate the optimal value.
 
         Args:
             term (str): e.g., ``"Length(1)"``
@@ -777,8 +826,7 @@ class Theory(object):
 
         result = solver.check([z3_form for z3_form, (_, expr) in ps.items() if
                                     not (expr and expr.code in self.ignored_laws)])
-        assert result == unsat, ("Incorrect solver result during explain inference. "
-                                 "This may be due to floating point imprecision.")
+        assert result == unsat, ("Theory is satisfiable: nothing to explain.")
         unsatcore = solver.unsat_core()
 
         solver.pop()
@@ -870,11 +918,11 @@ class Theory(object):
 
     def decision_table(self,
                        goal_string: str = "",
-                       timeout: int = 20,
+                       timeout_seconds: int = 20,
                        max_rows: int = 50,
                        first_hit: bool = True,
                        verify: bool = False
-                       ) -> List[List[Assignment]]:
+                       ) -> (List[List[Assignment]], bool):
         """Experimental.  Returns the rows for a decision table that defines ``goal_string``.
 
         ``goal_string`` must be a predicate application defined in the theory.
@@ -882,15 +930,17 @@ class Theory(object):
 
         Args:
             goal_string (str, optional): the last column of the table.
-            timeout (int, optional): maximum duration in seconds. Defaults to 20.
+            timeout_seconds (int, optional): maximum duration in seconds. Defaults to 20.
             max_rows (int, optional): maximum number of rows. Defaults to 50.
             first_hit (bool, optional): requested hit-policy. Defaults to True.
             verify (bool, optional): request verification of table completeness.  Defaults to False
 
         Returns:
             list(list(Assignment)): the non-empty cells of the decision table  for ``goal_string``, given ``self``.
+            bool: whether or not the timeout limit was reached.
         """
-        max_time = time.time()+timeout  # 20 seconds max
+        timeout_hit = False
+        max_time = time.time()+timeout_seconds  # 20 seconds max
         assert self.extended == True, \
             "The problem must be created with 'extended=True' for decision_table."
 
@@ -976,6 +1026,8 @@ class Theory(object):
             solver.add(modelZ3)
 
             count +=1
+        if time.time() > max_time:
+            timeout_hit = True
 
         if verify:
             def verify_models(known, models, goal_string):
@@ -1068,10 +1120,12 @@ class Theory(object):
                         new = new[0].negate()
                         models[i-1] = [new, models[i-1][1]]
                         del models[i]
+            if time.time() > max_time:
+                timeout_hit = True
             if verify:
                 verify_models(known, models, goal_string)
 
-        return models
+        return (models, timeout_hit)
 
     def _first_propagate(self, solver):
         """monkey-patched"""
