@@ -6,6 +6,7 @@ from .Simplify import Done
 from .Idp_to_Z3 import Done
 from .Run import (model_check, model_expand, model_propagate,
                   execute, decision_table)
-from .Problem import Problem
+from .Theory import Propagation, Theory
 from .Propagate import Done
-from .Assignments import Status
\ No newline at end of file
+from .Relevance import Done
+from .Assignments import Status, Assignment, Assignments
