@@ -19,7 +19,9 @@
     Various utilities (in particular, OrderedSet)
 """
 
-from collections import ChainMap, Iterable
+from collections import ChainMap
+from collections.abc import Iterable
+from datetime import datetime
 from json import JSONEncoder
 import time
 from enum import Enum, auto
@@ -48,22 +50,24 @@ DEF_SEMANTICS = Semantics.WELLFOUNDED
 
 NEWL = "\n"
 indented = "\n  "
-BOOL = "Bool"
-INT = "Int"
-REAL = "Real"
-DATE = "Date"
-SYMBOL = "Symbol"
 
-RELEVANT = "__relevant"
+BOOL = "𝔹"
+INT = "ℤ"
+REAL = "ℝ"
+DATE = "Date"
+CONCEPT = "Concept"
+
+GOAL_SYMBOL = "goal_symbol"
+RELEVANT = " relevant"  # internal.  Leading space to avoid conflict with user vocabulary
+EXPAND = "expand"
 ABS = "abs"
-ARITY = "arity"
-INPUT_DOMAIN = "input_domain"
-OUTPUT_DOMAIN = "output_domain"
-RESERVED_SYMBOLS = [BOOL, INT, REAL, DATE, SYMBOL,
-                    RELEVANT, ABS, ARITY, INPUT_DOMAIN, OUTPUT_DOMAIN]
+RESERVED_SYMBOLS = [BOOL, INT, REAL, DATE, CONCEPT,
+                    GOAL_SYMBOL, RELEVANT, ABS, ]
 
 DEFAULT = "default"
 
+NOT_SATISFIABLE ="Not satisfiable."
+
 """ Module that monkey-patches json module when it's imported so
 JSONEncoder.default() automatically checks for a special "to_json()"
 method and uses it to encode the object if found.
@@ -81,7 +85,7 @@ start = time.process_time()
 
 def log(action):
     global start
-    print("*** ", action, round(time.process_time()-start, 3))
+    print("*** ", action, datetime.now().strftime("%H:%M:%S"), round(time.process_time()-start, 3))
     start = time.process_time()
 
 
@@ -96,38 +100,6 @@ def unquote(s):
     return s
 
 
-def in_list(q, ls):
-    if not ls:
-        return True  # e.g. for INT, REAL, DATE
-    if len(ls) == 1:
-        return q == ls[0]
-    return Or([q == i for i in ls])
-
-
-def is_number(s):
-    if str(s) in ['True', 'False']:
-        return False
-    try:
-        float(eval(str(s if not s.endswith('?') else s[:-1]))) # accepts "2/3" or "3.1415?"
-        return True
-    except:
-        return False
-
-
-def splitLast(l):
-    return l[:-1], l[-1]
-
-
-def applyTo(sym, arg):
-    if len(arg) == 0:
-        return sym
-    return sym(arg)
-
-
-def mergeDicts(l):
-    # merge a list of dicts (possibly a comprehension
-    return dict(ChainMap(*reversed(list(l))))
-
 # OrderedSet  #############################################
 
 
